@@ -6,31 +6,44 @@
  */
 
 #include "include/core/SkDataTable.h"
+#include "include/core/SkFontArguments.h"
 #include "include/core/SkFontMgr.h"
 #include "include/core/SkFontStyle.h"
+#include "include/core/SkMatrix.h"
 #include "include/core/SkRefCnt.h"
+#include "include/core/SkScalar.h"
 #include "include/core/SkStream.h"
 #include "include/core/SkString.h"
 #include "include/core/SkTypeface.h"
 #include "include/core/SkTypes.h"
-#include "include/private/base/SkFixed.h"
-#include "include/private/base/SkMath.h"
+#include "include/ports/SkFontMgr_fontconfig.h"
+#include "include/private/base/SkDebug.h"
 #include "include/private/base/SkMutex.h"
+#include "include/private/base/SkTArray.h"
 #include "include/private/base/SkTDArray.h"
 #include "include/private/base/SkTemplates.h"
+#include "include/private/base/SkThreadAnnotations.h"
+#include "src/base/SkTSort.h"
 #include "src/core/SkAdvancedTypefaceMetrics.h"
 #include "src/core/SkFontDescriptor.h"
 #include "src/core/SkOSFile.h"
+#include "src/core/SkScalerContext.h"
 #include "src/core/SkTypefaceCache.h"
-#include "src/ports/SkFontHost_FreeType_common.h"
+#include "src/ports/SkTypeface_FreeType.h"
 
 #include <fontconfig/fontconfig.h>
-#include <string.h>
 
-using namespace skia_private;
+#include <string.h>
+#include <array>
+#include <cstddef>
+#include <cstdint>
+#include <memory>
+#include <utility>
 
 class SkData;
 
+using namespace skia_private;
+
 // FC_POSTSCRIPT_NAME was added with b561ff20 which ended up in 2.10.92
 // Ubuntu 14.04 is on 2.11.0
 // Debian 8 and 9 are on 2.11
@@ -480,7 +493,8 @@ public:
     }
 
     sk_sp<SkTypeface> onMakeClone(const SkFontArguments& args) const override {
-        std::unique_ptr<SkFontData> data = this->cloneFontData(args);
+        SkFontStyle style = this->fontStyle();
+        std::unique_ptr<SkFontData> data = this->cloneFontData(args, &style);
         if (!data) {
             return nullptr;
         }
@@ -489,7 +503,7 @@ public:
         SkString familyName;
         this->getFamilyName(&familyName);
         return sk_make_sp<SkTypeface_FreeTypeStream>(
-            std::move(data), familyName, this->fontStyle(), this->isFixedPitch());
+            std::move(data), familyName, style, this->isFixedPitch());
     }
 
     std::unique_ptr<SkFontData> onMakeFontData() const override {
@@ -523,12 +537,12 @@ class SkFontMgr_fontconfig : public SkFontMgr {
     mutable SkAutoFcConfig fFC;  // Only mutable to avoid const cast when passed to FontConfig API.
     const SkString fSysroot;
     const sk_sp<SkDataTable> fFamilyNames;
-    const SkTypeface_FreeType::Scanner fScanner;
 
     class StyleSet : public SkFontStyleSet {
     public:
         StyleSet(sk_sp<SkFontMgr_fontconfig> parent, SkAutoFcFontSet fontSet)
-            : fFontMgr(std::move(parent)), fFontSet(std::move(fontSet))
+            : fFontMgr(std::move(parent))
+            , fFontSet(std::move(fontSet))
         { }
 
         ~StyleSet() override {
@@ -702,33 +716,44 @@ protected:
     /** True if any string object value in the font is the same
      *         as a string object value in the pattern.
      */
-    static bool AnyMatching(FcPattern* font, FcPattern* pattern, const char* object) {
-        FcChar8* fontString;
-        FcChar8* patternString;
-        FcResult result;
-        // Set an arbitrary limit on the number of pattern object values to consider.
-        // TODO: re-write this to avoid N*M
-        static const int maxId = 16;
-        for (int patternId = 0; patternId < maxId; ++patternId) {
-            result = FcPatternGetString(pattern, object, patternId, &patternString);
-            if (FcResultNoId == result) {
-                break;
-            }
-            if (FcResultMatch != result) {
-                continue;
-            }
-            for (int fontId = 0; fontId < maxId; ++fontId) {
-                result = FcPatternGetString(font, object, fontId, &fontString);
-                if (FcResultNoId == result) {
+    static bool AnyStringMatching(FcPattern* font, FcPattern* pattern, const char* object) {
+        auto getStrings = [](FcPattern* p, const char* object, STArray<32, FcChar8*>& strings) {
+            // Set an arbitrary (but high) limit on the number of pattern object values to consider.
+            static constexpr const int maxId = 65536;
+            for (int patternId = 0; patternId < maxId; ++patternId) {
+                FcChar8* patternString;
+                FcResult result = FcPatternGetString(p, object, patternId, &patternString);
+                if (result == FcResultNoId) {
                     break;
                 }
-                if (FcResultMatch != result) {
-                    continue;
-                }
-                if (0 == FcStrCmpIgnoreCase(patternString, fontString)) {
-                    return true;
+                if (result == FcResultMatch) {
+                    strings.push_back(patternString);
                 }
             }
+        };
+        auto compareStrings = [](FcChar8* a, FcChar8* b) -> bool {
+            return FcStrCmpIgnoreCase(a, b) < 0;
+        };
+
+        STArray<32, FcChar8*> fontStrings;
+        STArray<32, FcChar8*> patternStrings;
+        getStrings(font, object, fontStrings);
+        getStrings(pattern, object, patternStrings);
+
+        SkTQSort(fontStrings.begin(), fontStrings.end(), compareStrings);
+        SkTQSort(patternStrings.begin(), patternStrings.end(), compareStrings);
+
+        FcChar8** fontString = fontStrings.begin();
+        FcChar8** patternString = patternStrings.begin();
+        while (fontString != fontStrings.end() && patternString != patternStrings.end()) {
+            int cmp = FcStrCmpIgnoreCase(*fontString, *patternString);
+            if (cmp < 0) {
+                ++fontString;
+            } else if (cmp > 0) {
+                ++patternString;
+            } else {
+                return true;
+            }
         }
         return false;
     }
@@ -761,7 +786,7 @@ protected:
     }
 
     static bool FontFamilyNameMatches(FcPattern* font, FcPattern* pattern) {
-        return AnyMatching(font, pattern, FC_FAMILY);
+        return AnyStringMatching(font, pattern, FC_FAMILY);
     }
 
     static bool FontContainsCharacter(FcPattern* font, uint32_t character) {
@@ -789,7 +814,7 @@ protected:
         FCLocker lock;
 
         SkAutoFcPattern pattern;
-        FcPatternAddString(pattern, FC_FAMILY, (FcChar8*)familyName);
+        FcPatternAddString(pattern, FC_FAMILY, (const FcChar8*)familyName);
         FcConfigSubstitute(fFC, pattern, FcMatchPattern);
         FcDefaultSubstitute(pattern);
 
@@ -833,7 +858,7 @@ protected:
             FCLocker lock;
 
             SkAutoFcPattern pattern;
-            FcPatternAddString(pattern, FC_FAMILY, (FcChar8*)familyName);
+            FcPatternAddString(pattern, FC_FAMILY, (const FcChar8*)familyName);
             fcpattern_from_skfontstyle(style, pattern);
             FcConfigSubstitute(fFC, pattern, FcMatchPattern);
             FcDefaultSubstitute(pattern);
@@ -943,6 +968,6 @@ protected:
     }
 };
 
-SK_API sk_sp<SkFontMgr> SkFontMgr_New_FontConfig(FcConfig* fc) {
+sk_sp<SkFontMgr> SkFontMgr_New_FontConfig(FcConfig* fc) {
     return sk_make_sp<SkFontMgr_fontconfig>(fc);
 }
